from . import NBTHandler
from . import Utilities
from . import bitstack
import math

def Resolve(fPath:str):
    litematic = open(fPath, "rb")
    binSource = Utilities.GZipUnzip(litematic.read())
    return decode_BlockStates(to_human(NBTHandler.Resolve(binSource)))


def to_human(Resolve_data:dict) -> dict:
    for i in Resolve_data["Metadata"]["EnclosingSize"]:
        Resolve_data["Metadata"]["EnclosingSize"][i] = str(
            int(int(Resolve_data["Metadata"]["EnclosingSize"][i]) / 16777216)
        )
    for i in Resolve_data["Regions"]:
        for y in Resolve_data["Regions"][i]["Size"]:
            Resolve_data["Regions"][i]["Size"][y] = str(
                math.floor(float(Resolve_data["Regions"][i]["Size"][y]) / 16777216)
            )

        for y in Resolve_data["Regions"][i]["Position"]:
            Resolve_data["Regions"][i]["Position"][y] = str(
                int(int(Resolve_data["Regions"][i]["Position"][y]) / 16777216)
            )
        for y, z in enumerate(Resolve_data["Regions"][i]["TileEntities"]):
            for hh in Resolve_data["Regions"][i]["TileEntities"][y]:
                if hh == "x" or hh == "y" or hh == "z":
                    Resolve_data["Regions"][i]["TileEntities"][y][hh] = str(
                        int(int(Resolve_data["Regions"][i]["TileEntities"][y][hh]) / 16777216)
                    )

    return Resolve_data


def decode_BlockStates(Resolve_data:dict) -> dict:
    #   "BlockStates":[
    #      "360323773641625836",
    #      "977288815838265344"
    #   ]
    #     (palette entries 0x00 through 0x1d are valid):

    # 0x00 - air
    # 0x01 - iron_block
    # 0x02 - white_concrete
    # (first couple of words of BlockState, spaces added for clarity):

    # 0x 1084 4201 8410 8400
    # 0x 4210 8420 8400 0842

    # Grouping the first word's binary data little-endian:
    # (last) (0001) 00001 00001 00010 00010 00010 00010 00010 00001 00001 00001 00000 00000 (first)

    for i in Resolve_data["Regions"]:
        Resolve_data["Regions"][i]["decode_BlockStates"] = []
        bitlist = bitstack.bitstack(
            len(Resolve_data["Regions"][i]["BlockStatePalette"]),
            Resolve_data["Regions"][i]["BlockStatePalette"],
        )
        for y in Resolve_data["Regions"][i]["BlockStates"]:
            bitlist.add(y)
        Resolve_data["Regions"][i]["decode_BlockStates"] = bitlist.calc()
        print("decode success")
    return Resolve_data



import struct
from . import Utilities
import copy
from typing import Union


def Resolve(binSource:bytes , indentationWhiteSpace:int = 2) -> dict:
    jsonLines , _ = ReadCompoundTag(binSource , indentationWhiteSpace)
    return jsonLines

def ReadCompoundTag(binSource:bytes , indentationWhiteSpace:int = 2 , pointer:int = 0 , defaultIndentation:int = 0 , readName:bool = False) -> tuple[Union[dict , str] , int]:
    indentation = defaultIndentation
    name = ""
    if readName:
        name , pointer = ReadTagName(binSource , pointer)
        litematicdata = {}
        returnlitematicdata = {"name": name, "value": ""}
    else:
        pointer += 3
        litematicdata = {}

    while True:
        if binSource[pointer] == 0:
            indentation -= indentationWhiteSpace
            pointer += 1
            if indentation < defaultIndentation:
                break
        if   binSource[pointer] == 1:  # TAG_Byte
            jsonLine , pointer = ReadByteTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 2:  # TAG_Short
            jsonLine , pointer = ReadShortTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 3:  # TAG_Int
            jsonLine , pointer = ReadIntTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 4:  # TAG_Long
            jsonLine , pointer = ReadLongTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 5:  # TAG_Float
            jsonLine , pointer = ReadFloatTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 6:  # TAG_Double
            jsonLine , pointer = ReadDoubleTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 7:  # TAG_Byte_Array
            jsonSubLines , pointer = ReadByteArrayTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonSubLines["name"]] = jsonSubLines["value"]
        elif binSource[pointer] == 8:  # TAG_String
            jsonLine , pointer = ReadStringTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonLine["name"]] = jsonLine["value"]
        elif binSource[pointer] == 9:  # TAG_List
            jsonSubLines , pointer = ReadListTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonSubLines["name"]] = jsonSubLines["value"]
        elif binSource[pointer] == 10: # TAG_Compound
            jsonSubLines , pointer = ReadCompoundTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2 , True)
            litematicdata[jsonSubLines["name"]] = jsonSubLines["value"]
        elif binSource[pointer] == 11: # TAG_Int_Array
            jsonSubLines , pointer = ReadIntArrayTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonSubLines["name"]] = jsonSubLines["value"]
        elif binSource[pointer] == 12: # TAG_Long_Array
            jsonSubLines , pointer = ReadLongArrayTag(copy.deepcopy(binSource) , indentationWhiteSpace , pointer , indentation + 2)
            litematicdata[jsonSubLines["name"]] = jsonSubLines["value"]
        else:
            break
    if readName:
        returnlitematicdata["value"] = litematicdata
        return returnlitematicdata , pointer
    else:
        return litematicdata , pointer


def ReadByteTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent , pointer = ReadByte(binSource , pointer)
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer


def ReadShortTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent = str(struct.unpack("h" , binSource[pointer : pointer + 2]))[1 : -2]
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer + 2


def ReadIntTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent = str(struct.unpack("i" , binSource[pointer : pointer + 4]))[1 : -2]
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer + 4


def ReadLongTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent = str(struct.unpack("q" , binSource[pointer : pointer + 8]))[1 : -2]
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer + 8


def ReadFloatTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent = ReadFloat(binSource , pointer)
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer + 4


def ReadDoubleTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent = ReadDouble(binSource , pointer)
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer + 8


def ReadByteArrayTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    length = Utilities.BigEndiannessForInt(binSource , pointer)
    jsonLines = {"name": tagName, "value": []}
    pointer += 4
    for _ in range(length):
        tagContent , pointer = ReadByte(binSource , pointer)
        jsonLines["value"].append(tagContent)

    return jsonLines , pointer


def ReadStringTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    tagContent , pointer = ReadString(binSource , pointer + 1)
    jsonLine = {"name": tagName, "value": tagContent}
    return jsonLine , pointer


def ReadListTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation , callByList = False):
    if callByList:
        tagName = "SubList"
    else:
        tagName , pointer = ReadTagName(binSource , pointer)
    jsonLines = {"name": tagName, "value": []}
    contentType = binSource[pointer]
    contentCount = Utilities.BigEndiannessForInt(binSource , pointer + 1)
    pointer += 5
    if contentType != 0:
        if contentType == 1:    # TAG_Bytes
            for _ in range(contentCount):
                tagContent , pointer = ReadByte(binSource , pointer)
                jsonLines["value"].append(tagContent)
        elif 1 < contentType < 5: #TAG_Short , TAG_Int , TAG_Long
            unpackFormat = None
            if   contentType == 2:
                unpackFormat = "h"
            elif contentType == 3:
                unpackFormat = "i"
            else:
                unpackFormat = "q"
            size = (1 << contentType) >> 1
            for _ in range(contentCount):
                tagContent = str(struct.unpack(unpackFormat , binSource[pointer : pointer + size]))[1 : -2]
                jsonLines["value"].append(tagContent)
        elif contentType == 5: # TAG_Float
            for _ in range(contentCount):
                tagContent , pointer = ReadFloat(binSource , pointer)
                jsonLines["value"].append(tagContent)
        elif contentType == 6: # TAG_Double
            for _ in range(contentCount):
                tagContent , pointer = ReadDouble(binSource , pointer)
                jsonLines["value"].append(tagContent)
        elif contentType == 8:    # TAG_String
            for _ in range(contentCount):
                pointer += 1
                tagContent , pointer = ReadString(binSource , pointer)
                jsonLines["value"].append(tagContent)
        elif contentType == 9:    # TAG_List
            for _ in range(contentCount):
                jsonSubLines , pointer = ReadListTag(binSource , indentationWhiteSpace , pointer , defaultIndentation + 2 , True)
                jsonLines["value"].append(jsonSubLines)
        elif contentType == 10:   # TAG_Compound
            for _ in range(contentCount):
                jsonSubLines , pointer = ReadCompoundTag(binSource , indentationWhiteSpace , pointer - 3 , defaultIndentation + 2 , False)
                jsonLines["value"].append(jsonSubLines)
        elif contentType == 7 or 10 < binSource[pointer] < 13:  # TAG_Byte_Array , TAG_Int_Array , TAG_Long_Array
            pass
    return jsonLines , pointer


def ReadIntArrayTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    length = Utilities.BigEndiannessForInt(binSource , pointer)
    jsonLines = {"name": tagName, "value": []}
    pointer += 4
    for _ in range(length):
        binSource = binSource.ljust(pointer + 4 , b"\x00")
        tagContent = str(struct.unpack("i" , binSource[pointer : pointer + 4]))[1 : -2]
        jsonLines["value"].append(tagContent)
        pointer += 4
    return jsonLines , pointer


def ReadLongArrayTag(binSource:bytes , indentationWhiteSpace , pointer , defaultIndentation):
    tagName , pointer = ReadTagName(binSource , pointer)
    length = Utilities.BigEndiannessForInt(binSource , pointer)
    jsonLines = {"name": tagName, "value": []}
    pointer += 4
    for _ in range(length):
        tagContent = str(struct.unpack("q" , binSource[pointer : pointer + 8]))[1 : -2]
        jsonLines["value"].append(tagContent)
        pointer += 8

    return jsonLines , pointer


def ReadString(binSource:bytes , pointer):
    string = ""
    for _ in range(binSource[pointer]):
        pointer += 1
        string += chr(binSource[pointer])
    return string , pointer + 1


def ReadFloat(binSource:bytes , pointer):
    return float(str(struct.unpack(">f" , bytearray(binSource[pointer : pointer + 4])))[1:-2]) , pointer + 4


def ReadDouble(binSource:bytes , pointer):
    return float(str(struct.unpack(">d" , binSource[pointer : pointer + 8]))[1:-2]) , pointer + 8


def ReadByte(binSource:bytes , pointer):
    binary = binSource[pointer]
    if (binary & 0b1000_0000) > 0:
        binary = -((~binary + 1) & 0b1111_1111)
    return binary , pointer + 1


def ReadTagName(binSource:bytes , pointer):
    pointer += 2
    if binSource[pointer] == 0:
        return "Unknow Name" , pointer + 1
    return ReadString(binSource , pointer)


import json
import os
import shutil

class convert_texturepack:

    def __init__(self, path: str , output: str):
        self.path = path
        self.mainpath = os.path.join(path, "assets", "minecraft")
        self.output = output
        os.makedirs(self.output, exist_ok=True)
        self.noneedfind = ["armor_trims","mob_effects","shield_patterns","particles"]
        self.blocksdata = {"models":{}}
        self.start()


    def start(self) -> None:
        sources = []
        for i in os.listdir(os.path.join(self.mainpath, "atlases")):
            if i.split(".")[0] in self.noneedfind:
                continue
            with open(os.path.join(self.mainpath, "atlases", i), "r", encoding="utf8") as f:
                blocksdata = json.load(f)
            for j in blocksdata["sources"]:
                if j["type"] == "paletted_permutations":
                    continue
                sources.append(j)
        needcopy = set()
        for i in sources:
            if i['type'] == "directory":
                needcopy.add(i["source"].split("/")[0])
            elif i['type'] == "single":
                needcopy.add(i["resource"].split("/")[0])

        def load_model(path:str) -> None:
            with open(os.path.join(self.mainpath, "models", path), "r", encoding="utf8") as f:
                blockmodel = f.read()
                tempload = json.loads(blockmodel)
            blockmodel = blockmodel.replace("minecraft:", "")
            blockmodel:dict = json.loads(blockmodel)
            if "parent" in blockmodel and blockmodel["parent"]  not in self.blocksdata["models"]:
                load_model(tempload["parent"].split(":")[-1]+".json")
            # 強制複寫無法解決問題的UV
            if path not in self.blocksdata["models"]:
                if path.split("/")[-1].split(".")[0] == "sculk_sensor":
                    blockmodel["elements"][0]["faces"]["north"]["uv"] = [0, 0, 16, 8]
                    blockmodel["elements"][0]["faces"]["east"]["uv"] = [0, 0, 16, 8]
                    blockmodel["elements"][0]["faces"]["south"]["uv"] = [0, 0, 16, 8]
                    blockmodel["elements"][0]["faces"]["west"]["uv"] = [0, 0, 16, 8]
                self.blocksdata["models"][path.split("/")[-1].split(".")[0]] = blockmodel

        for i in os.listdir(os.path.join(self.mainpath, "blockstates")):
            with open(os.path.join(self.mainpath, "blockstates", i), "r", encoding="utf8") as f:
                blockstates = json.load(f)
            if "variants" in blockstates:
                for variants in blockstates["variants"]:
                    if isinstance(blockstates["variants"][variants] , dict):
                        load_model(blockstates["variants"][variants]["model"].split(":")[-1]+".json")
                        blockstates["variants"][variants]["model"] = blockstates["variants"][variants]["model"].split("/")[-1]
                    elif isinstance(blockstates["variants"][variants] , list):
                        for j in blockstates["variants"][variants]:
                            load_model(j["model"].split(":")[-1]+".json")
                            j["model"] = j["model"].split("/")[-1]
            if "multipart" in blockstates:
                for multipart in blockstates["multipart"]:
                    if isinstance(multipart["apply"], dict):
                        load_model(multipart["apply"]["model"].split(":")[-1]+".json")
                        multipart["apply"]["model"] = multipart["apply"]["model"].split("/")[-1]
                    elif isinstance(multipart["apply"], list):
                        for j in multipart["apply"]:
                            load_model(j["model"].split(":")[-1]+".json")
                            j["model"] = j["model"].split("/")[-1]

            self.blocksdata[i.split(".")[0]] = blockstates

        with open(self.output+r"\output.json", "w", encoding="utf8") as f:
            json.dump(self.blocksdata, f, indent=4, ensure_ascii=False)

        if os.path.exists(os.path.join(self.output,"textures")):
            shutil.rmtree(os.path.join(self.output,"textures"))
        for i in needcopy:
            shutil.copytree(os.path.join(self.mainpath, "textures", i), os.path.join(self.output, "textures", i))





if __name__ == "__main__":
    convert_texturepack(r"C:\Users\phill\OneDrive\桌面\codetool\VanillaDefault+1.20", r"C:\Users\phill\OneDrive\Documents\coed_thing\3Dlitematica\temp")


import os
from pathlib import Path
import click
import json
from alive_progress import alive_bar
from .litematicadecoder import Resolve
from .objbuilder import LitimaticaToObj
from .texturepackexport import convert_texturepack


class Litematica(click.ParamType):
    name = "litematica"

    def convert(self, value, param, ctx):
        if not os.path.exists(value):
            self.fail(f"{value} does not exist.")
        if not value.endswith(".litematic"):
            self.fail(f"{value} is not a litematica file.")
        return value


class LitematicaOrJson(click.ParamType):
    name = "LitematicaOrJson"

    def convert(self, value, param, ctx):
        if not os.path.exists(value):
            self.fail(f"{value} does not exist.")
        if not value.endswith(".litematic") and not value.endswith(".json"):
            self.fail(f"{value} is not a litematica or json file.")
        return value


@click.group()
@click.option("--debug", default=False)
def cli(debug):
    if debug:
        click.echo("Debug mode is 'on' ")


@cli.command()
@click.argument("litematica", type=Litematica())
@click.option("-o", "--output", "output", default="./", help="Output file path")
@click.option("-f", "--filename", "filename", default="output.json", help="Output file name")
def Decode(litematica, output, filename):
    """
    Decode a litematica file to json file
    """
    path = Path(output).absolute()
    with alive_bar(bar="bubbles", spinner="wait"):
        data = Resolve(litematica)
    with open(os.path.join(path,filename), "w", encoding="utf8") as f:
        json.dump(data, f, indent=4)


@cli.command()
@click.argument("json_or_litematica", type=LitematicaOrJson())
@click.argument("texturefolder", type=click.Path(exists=True))
@click.option("-o", "--output", "output", default="./", help="Output file path")
def Obj(json_or_litematica, texturefolder, output):
    """
    Convert a litematica file to obj file
    """
    json_or_litematica = Path(json_or_litematica).absolute()
    TextureFolder = Path(texturefolder).absolute()
    output = Path(output).absolute()
    with alive_bar(bar="bubbles", spinner="wait"):
        if str(json_or_litematica).endswith(".litematic"):
            litematica = Resolve(json_or_litematica)
        else:
            print(json_or_litematica)
            with open(json_or_litematica, "r", encoding="utf8") as f:
                litematica = json.load(f)
        LitimaticaToObj(litematica, TextureFolder, output)


@cli.command()
@click.argument("texturepack", type=click.Path(exists=True))
@click.option("-o", "--output", "output", default="./temp", help="Output file path")
def Texture(texturepack, output):
    """
    Convert texture pack for 3d litematica
    """
    texturepack = Path(texturepack).absolute()
    output = Path(output).absolute()
    with alive_bar(bar="bubbles", spinner="wait"):
        convert_texturepack(texturepack, output)


if __name__ == "__main__":
    cli()



from typing import List, Any
class bitstack:
    def __init__(self, bytelong:int, Resolve_data: List[Any]) -> None:
        self.bites = ""
        long = 1
        while bytelong > 2**long:
            long += 1
        self.bytelong = long
        self.Resolve_data = Resolve_data

    def add(self, bite: int) -> None:
        bite = int(bite)
        bite = bin(bite & 0xFFFFFFFFFFFFFFFF)[2:].zfill(64)
        bite = [j for j in [bite[i : i + 8] for i in range(0, len(bite), 8)][::-1]]
        bite = "".join(bite)
        self.bites += bite[::-1]

    def get(self, length: int) -> str:
        bite = self.bites[:length]
        self.bites = self.bites[length:]
        return bite

    def calc(self) -> List[Any]:
        stepbytes = self.bites
        stepbytes = [
            stepbytes[i : i + self.bytelong][::-1] for i in range(0, len(self.bites), self.bytelong)
        ]  # self.bytelong位一組
        stepbytes = stepbytes[::-1]
        stepbytes = [int(i, 2) for i in stepbytes]  # 二進位轉十進位
        decode_BlockStates = []
        for z in stepbytes:
            decode_BlockStates.append(self.Resolve_data[z])
        return decode_BlockStates



import json
import math
import os
from typing import Callable, List, Union
# example:
#         "cube": {
#             "parent": "block",
#             "elements": [
#                 {
#                     "from": [
#                         0,
#                         0,
#                         0
#                     ],
#                     "to": [
#                         16,
#                         16,
#                         16
#                     ],
#                     "faces": {
#                         "down": {
#                             "texture": "#down",
#                             "cullface": "down"
#                         },
#                         "up": {
#                             "texture": "#up",
#                             "cullface": "up"
#                         },
#                         "north": {
#                             "texture": "#north",
#                             "cullface": "north"
#                         },
#                         "south": {
#                             "texture": "#south",
#                             "cullface": "south"
#                         },
#                         "west": {
#                             "texture": "#west",
#                             "cullface": "west"
#                         },
#                         "east": {
#                             "texture": "#east",
#                             "cullface": "east"
#                         }
#                     }
#                 }
#             ]
#         },


# element 裡面的 from , to 是指一個block的3維空間的座標


class Enity:
    def __init__(self, x:float, y:float, z:float, blockdata:dict,texturepath:str) -> None:
        self.blockdata = blockdata
        self.name = blockdata["Name"].replace("minecraft:", "")
        self.x = x
        self.y = y
        self.z = z
        self.center = [sum([x, x + 0.1]) / 2, sum([y, y + 0.1]) / 2, sum([z, z + 0.1]) / 2]
        self.thisdata = None
        self.textures = {}
        self.enitys = []
        self.rotatemode = []
        self.rotate = []
        self.element = None
        self.objdata = {"blockname": self.name, "v": [], "vt": [], "f": [], "textures": []}
        self.texturepath = os.path.join(texturepath, "output.json")
        self.parse()
        self.merge()

    def merge(self) -> None:
        # f要替換為全部v轉換的index
        for i in self.enitys:
            self.objdata["v"].extend(i.objdata["v"])

        for i in self.enitys:
            for j in i.objdata["f"]:
                temp = []
                for x in range(len(j)):
                    temp.append(self.objdata["v"].index(i.objdata["v"][j[x] - 1]) + 1)
                self.objdata["f"].append(temp)
            self.objdata["vt"].extend(i.objdata["vt"])
            self.objdata["textures"].extend(i.objdata["textures"])

    def parse(self) -> None:
        with open(
            self.texturepath,
            "r",
            encoding="utf8",
        ) as f:
            self.thisdata = json.load(f)[self.name]
        if "variants" in self.thisdata:
            for i in self.thisdata["variants"]:
                if i != "":
                    flag = False
                    values = i.split(",")
                    for value in values:
                        value = value.split("=")
                        if value[0] in self.blockdata["Properties"]:
                            if value[1] == self.blockdata["Properties"][value[0]]:
                                flag = True
                            else:
                                flag = False
                                break
                    if flag:
                        if "x" in self.thisdata["variants"][i]:
                            self.rotatemode.append("x")
                            self.rotate.append(self.thisdata["variants"][i]["x"])
                        if "y" in self.thisdata["variants"][i]:
                            self.rotatemode.append("y")
                            self.rotate.append(self.thisdata["variants"][i]["y"])
                        self.load_model(self.thisdata["variants"][i]["model"])
                else:
                    self.load_model(self.thisdata["variants"][i]["model"])
        elif "multipart" in self.thisdata:
            code = 0
            for i in self.thisdata["multipart"]:
                if "when" in i:
                    if "OR" in i["when"]:
                        flag = False
                        for j in i["when"]["OR"]:
                            flag2 = False
                            for x in j.keys():
                                if "|" in j[x]:
                                    if self.blockdata["Properties"][x] in j[x].split("|"):
                                        flag2 = True
                                    else:
                                        flag2 = False
                                        break
                                else:
                                    if self.blockdata["Properties"][x] == j[x]:
                                        flag2 = True
                                    else:
                                        flag2 = False
                                        break
                            if flag2:
                                flag = True
                        if flag:
                            if "x" in i["apply"]:
                                self.rotatemode.append({"code": code, "mode": "x"})
                                self.rotate.append(i["apply"]["x"])
                            if "y" in i["apply"]:
                                self.rotatemode.append({"code": code, "mode": "y"})
                                self.rotate.append(i["apply"]["y"])
                            self.load_model(i["apply"]["model"].split("/")[-1], code)
                    else:
                        if (
                            self.blockdata["Properties"][list(i["when"].keys())[0]]
                            in list(i["when"].values())[0].split("|")
                            if "|" in list(i["when"].values())[0]
                            else self.blockdata["Properties"][list(i["when"].keys())[0]]
                            == list(i["when"].values())[0]
                        ):
                            if "x" in i["apply"]:
                                self.rotatemode.append({"code": code, "mode": "x"})
                                self.rotate.append(i["apply"]["x"])
                            if "y" in i["apply"]:
                                self.rotatemode.append({"code": code, "mode": "y"})
                                self.rotate.append(i["apply"]["y"])
                            self.load_model(i["apply"]["model"].split("/")[-1], code)
                else:
                    self.load_model(i["apply"]["model"].split("/")[-1], code)
                code += 1

    def load_model(self, modelname:str, code=None,isparent=False) -> None:
        with open(
            self.texturepath,
            "r",
            encoding="utf8",
        ) as f:
            model = json.load(f)["models"][modelname]
            if "parent" in model:
                self.load_model(model["parent"].split("/")[-1], code,True)
            # if isparent:
            #     self.parents[modelname] = model
            if "textures" in model:
                self.textures.update(model["textures"])
            if "elements" in model:
                self.element = model["elements"]

        if not isparent:
            self.enitys.append(Build_enity(self, code,self.element))


class Build_enity:
    def __init__(self, mother: Enity, code:str,elements:List[dict]) -> None:
        self.name = code
        self.x = mother.x
        self.y = mother.y
        self.z = mother.z
        self.center = [
            sum([self.x, self.x + 0.1]) / 2,
            sum([self.y, self.y + 0.1]) / 2,
            sum([self.z, self.z + 0.1]) / 2,
        ]
        self.parents = {}
        self.textures = mother.textures
        self.element = elements
        self.objdata = {"v": [], "vt": [], "f": [], "textures": []}
        self.rotatemode = mother.rotatemode
        self.rotate = mother.rotate
        self.start()

    def start(self):
        for i in self.element:
            self.build_element(i)
        alrrote = []

        for i in range(len(self.rotatemode)):
            if isinstance(self.rotatemode[i],dict):
                if self.rotatemode[i]["mode"] == "y" and self.rotatemode[i] not in alrrote and self.rotatemode[i]["code"] == self.name:
                    alrrote.append(self.rotatemode[i])
                    for j in range(len(self.objdata["v"])):
                        self.objdata["v"][j] = self.rotate_y(
                            self.objdata["v"][j], self.rotate[i], self.center
                        )
                elif self.rotatemode[i]["mode"] == "x" and self.rotatemode[i] not in alrrote and self.rotatemode[i]["code"] == self.name:
                    alrrote.append(self.rotatemode[i])
                    for j in range(len(self.objdata["v"])):
                        self.objdata["v"][j] = self.rotate_x(
                            self.objdata["v"][j], self.rotate[i], self.center
                        )
            else:
                if self.rotatemode[i] == "y" and self.rotatemode[i] not in alrrote:
                    alrrote.append(self.rotatemode[i])
                    for j in range(len(self.objdata["v"])):
                        self.objdata["v"][j] = self.rotate_y(
                            self.objdata["v"][j], self.rotate[i], self.center
                        )
                elif self.rotatemode[i] == "x" and self.rotatemode[i] not in alrrote:
                    alrrote.append(self.rotatemode[i])
                    for j in range(len(self.objdata["v"])):
                        self.objdata["v"][j] = self.rotate_x(
                            self.objdata["v"][j], self.rotate[i], self.center
                        )

    def append_pos(self, thelist, item):
        """
        append 到 list並回傳index
        """
        if item not in thelist:
            thelist.append(item)
            return len(thelist)
        else:
            return thelist.index(item) + 1

    def add_texture(self, texturename):
        if texturename in self.textures:
            if "#" in self.textures[texturename]:
                temp = self.textures[texturename]
                temp = temp.replace("#", "")
                self.add_texture(temp)
            else:
                self.objdata["textures"].append(self.textures[texturename])
                return

    def add_F(self, listV:List[float], rotate:int=0, elerotate:Callable[[List[float]],List[float]]=None) -> None:
        f = []
        if rotate == 90:
            listV = [listV[3], listV[0], listV[1], listV[2]]
        if rotate == 180:
            listV = [listV[2], listV[3], listV[0], listV[1]]
        elif rotate == 270:
            listV = [listV[1], listV[2], listV[3], listV[0]]
        if elerotate != None:
            for i in range(len(listV)):
                listV[i] = elerotate(listV[i])
        for i in listV:
            f.append(self.append_pos(self.objdata["v"], i))
        self.objdata["f"].append(f)

    def build_element(self, element:dict) -> None:
        # 一個方塊 = 1*1*1
        # 數據給的方塊 = 16*16*16
        # 轉換為 1*1*1
        pos1 = element["from"]
        pos2 = element["to"]

        elerotate = None



        if "rotation" in element:
            center = []
            if element["rotation"]["angle"] > 0:
                element["rotation"]["angle"] = -element["rotation"]["angle"]
            else:
                element["rotation"]["angle"] = abs(element["rotation"]["angle"])
            if "origin" in element["rotation"]:
                center = [
                    element["rotation"]["origin"][i] / (16 * 10) + [self.x, self.y, self.z][i]
                    for i in range(3)
                ]
            if element["rotation"]["axis"] == "y":
                elerotate = lambda x: self.rotate_y(  # noqa: E731
                    x, element["rotation"]["angle"], center if center != [] else self.center
                )
            elif element["rotation"]["axis"] == "x":
                elerotate = lambda x: self.rotate_x(  # noqa: E731
                    x, element["rotation"]["angle"], center if center != [] else self.center
                )
            if "rescale" in element["rotation"] and element["rotation"]["rescale"] and "rail" in self.textures:
                pos1 = [0,9,-3]
                pos2 = [16,9,19]

        for i in range(3):
            pos1[i] /= 16 * 10
            pos2[i] /= 16 * 10



        # 六個面 = down up north south west east
        for i in element["faces"]:
            if "rotation" in element["faces"][i]:
                rotate = element["faces"][i]["rotation"]
            else:
                rotate = 0
            if i == "up":
                self.add_F(
                    [
                        [pos1[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                        [pos1[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])
            elif i == "down":
                self.add_F(
                    [
                        [pos1[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos2[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos2[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                        [pos1[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])
            elif i == "north":
                self.add_F(
                    [
                        [pos2[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos1[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos1[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])
            elif i == "south":
                self.add_F(
                    [
                        [pos1[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                        [pos2[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                        [pos1[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])
            elif i == "west":
                self.add_F(
                    [
                        [pos1[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos1[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                        [pos1[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                        [pos1[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])
            elif i == "east":
                self.add_F(
                    [
                        [pos2[0] + self.x, pos1[1] + self.y, pos2[2] + self.z],
                        [pos2[0] + self.x, pos1[1] + self.y, pos1[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos1[2] + self.z],
                        [pos2[0] + self.x, pos2[1] + self.y, pos2[2] + self.z],
                    ],
                    rotate,
                    elerotate,
                )
                if "texture" in element["faces"][i]:
                    self.add_texture(element["faces"][i]["texture"].replace("#", ""))
                else:
                    self.add_texture(i)
                self.add_vt(element["faces"][i])

    def rotate_y(self, point:List[float], angle:Union[int,float], center:List[float]):
        angle = math.radians(angle)
        x = point[0] - center[0]
        z = point[2] - center[2]
        point[0] = x * math.cos(angle) - z * math.sin(angle) + center[0]
        point[2] = x * math.sin(angle) + z * math.cos(angle) + center[2]
        return point

    def rotate_x(self, point:List[float], angle:Union[int,float], center:List[float]):
        angle = math.radians(angle)
        y = point[1] - center[1]
        z = point[2] - center[2]
        point[1] = y * math.cos(angle) + z * math.sin(angle) + center[1]
        point[2] = -y * math.sin(angle) + z * math.cos(angle) + center[2]
        return point

    def add_vt(self, face:dict) -> None:
        if "uv" in face:
            self.objdata["vt"].append(
                [
                    [
                        face["uv"][0] / 16 if face["uv"][0] != 0 else 0,
                        face["uv"][1] / 16 if face["uv"][1] != 0 else 0,
                    ],
                    [
                        face["uv"][2] / 16 if face["uv"][2] != 0 else 0,
                        face["uv"][1] / 16 if face["uv"][1] != 0 else 0,
                    ],
                    [
                        face["uv"][2] / 16 if face["uv"][2] != 0 else 0,
                        face["uv"][3] / 16 if face["uv"][3] != 0 else 0,
                    ],
                    [
                        face["uv"][0] / 16 if face["uv"][0] != 0 else 0,
                        face["uv"][3] / 16 if face["uv"][3] != 0 else 0,
                    ],
                ]
            )
        else:
            self.objdata["vt"].append(
                [
                    [0, 0],
                    [1, 0],
                    [1, 1],
                    [0, 1],
                ]
            )


if __name__ == "__main__":
    # Enity(0, 0, 0, {"Name": "minecraft:redstone_block"}).start()
    print(Build_enity.rotate_y(..., [10, 0, 2], 180, [5, 5, 5]))



import gzip
import io


def BigEndiannessForInt(source: bytes, index: int) -> int:
    return (
        (source[index] << 24)
        + (source[index + 1] << 16)
        + (source[index + 2] << 8)
        + source[index + 3]
    )


def BigEndiannessForLong(source: bytes, index: int) -> int:
    return (BigEndiannessForInt(source, index) << 32) + BigEndiannessForInt(source, index + 4)


def SmallEndiannessForInt(source: bytes, index: int) -> int:
    return (
        source[index]
        + (source[index + 1] << 8)
        + (source[index + 2] << 16)
        + (source[index + 3] << 24)
    )


def GZipUnzip(source:bytes) -> bytes:
    return gzip.GzipFile(fileobj=io.BytesIO(source)).read()



import json
import sys
import traceback
from .mctoobj import Enity
import os
from typing import List
import tempfile
import shutil
from pathlib import Path

def LitimaticaToObj(litematica: dict, TextureFolder: str, output: str = "./") -> None:
    size = (
        int(litematica["Metadata"]["EnclosingSize"]["x"]),
        int(litematica["Metadata"]["EnclosingSize"]["y"]),
        int(litematica["Metadata"]["EnclosingSize"]["z"]),
    )
    regonname = list(litematica["Regions"].keys())[0]
    litematica = litematica["Regions"][regonname]["decode_BlockStates"]
    Objhandel("test", litematica, size, TextureFolder, output)

class Objhandel:
    def __init__(self, name:str, data:List[dict], size:tuple[int,int,int],TextureFolder:str,outputfolder:str,show_error_block:bool=False) -> None:
        self.name = name
        self.tempfolder = tempfile.mkdtemp()
        self.objfile = open(os.path.join(self.tempfolder, self.name + ".obj"), "w")
        self.output = "# generate by 3dlitematica" + "\n" + "g " + name + "\n"
        self.tmpdata = {}
        self.vtof = {}  # 對應表
        self.vtovt = {}
        self.textures = []
        self.show_error_block = show_error_block
        self.TextureFolder = TextureFolder
        self.outputfolder = outputfolder
        self.main(data, size)
        shutil.make_archive(os.path.join(self.outputfolder, self.name), "zip", self.tempfolder)
        shutil.rmtree(self.tempfolder)

    def main(self, data, size):
        data = list(reversed(data))
        x = size[0]
        y = size[1]
        z = size[2]
        count = 0
        for j in range(1, y + 1):
            for k in range(1, z + 1):
                for i in range(1, x + 1):
                    if data[count]["Name"] == "minecraft:air":
                        pass
                    else:
                        try:
                            self.addEnity(Enity(i / 10, j / 10, k / 10, data[count], self.TextureFolder))
                        except Exception as e:
                            error_class = e.__class__.__name__
                            detail = e.args[0]
                            cl, exc, tb = sys.exc_info()
                            lastCallStack = traceback.extract_tb(tb)[-1]
                            fileName = lastCallStack[0]
                            lineNum = lastCallStack[1]
                            funcName = lastCallStack[2]
                            errMsg = 'File "{}", line {}, in {}: [{}] {}'.format(
                                fileName, lineNum, funcName, error_class, detail
                            )
                            print(f"[UserData] | {errMsg}")
                            if self.show_error_block:
                                self.addblock(i / 10, j / 10, k / 10, data[count]["Name"])
                            else:
                                pass
                    count += 1

        # self.objfile.write(self.output)
        self.writeobj()
        self.objfile.close()

    """
    # List of geometric vertices, with (x, y, z, [w]) coordinates, w is optional and defaults to 1.0.
    v 0.123 0.234 0.345 1.0
    v ...

    # List of texture coordinates, in (u, [v, w]) coordinates, these will vary between 0 and 1. v, w are optional and default to 0.
    vt 0.500 1 [0]
    vt ...

    # Polygonal face element
    f 1 2 3
    f 3/1 4/2 5/3
    f 6/4/1 3/5/3 7/6/5
    f 7//1 8//2 9//3
    f ...
    """

    def addEnity(self, Enity: Enity) -> None:
        if Enity.objdata["v"] == []:
            raise Exception('Enity.objdata["v"] is empty')
        self.tmpdata[(Enity.x, Enity.y, Enity.z)] = Enity.objdata

    def addblock(self, x:float, y:float, z:float, blockname:str) -> None:
        self.tmpdata[(x, y, z)] = {
            "blockname": blockname,
            "v": [
                [x, y, z],
                [x + 0.1, y, z],
                [x + 0.1, y, z + 0.1],
                [x, y, z + 0.1],
                [x, y + 0.1, z],
                [x + 0.1, y + 0.1, z],
                [x + 0.1, y + 0.1, z + 0.1],
                [x, y + 0.1, z + 0.1],
            ],
            "vt": [
                [[0, 0], [1, 0], [1, 1], [0, 1]],
                [[0, 0], [1, 0], [1, 1], [0, 1]],
                [[0, 0], [1, 0], [1, 1], [0, 1]],
                [[0, 0], [1, 0], [1, 1], [0, 1]],
                [[0, 0], [1, 0], [1, 1], [0, 1]],
                [[0, 0], [1, 0], [1, 1], [0, 1]],
            ],
            "f": [
                [1, 2, 3, 4],
                [8, 7, 6, 5],
                [5, 6, 2, 1],
                [6, 7, 3, 2],
                [7, 8, 4, 3],
                [8, 5, 1, 4],
            ],
        }

    def writeobj(self) -> None:
        # TODO: 將方塊加入到group裡面
        # https://blog.csdn.net/xyh930929/article/details/82260581
        # 格式 ：f v/vt/vn v/vt/vn v/vt/vn（f 顶点索引 / 纹理坐标索引 / 顶点法向量索引）
        temp = ""
        if not os.path.exists(os.path.join(self.tempfolder, "textures")):
            os.makedirs(os.path.join(self.tempfolder, "textures"))
        for i in self.tmpdata:
            if "textures" in self.tmpdata[i]:
                for j in set(self.tmpdata[i]["textures"]):
                    if j not in self.textures:
                        self.textures.append(j)
                        shutil.copy(
                            os.path.join(self.TextureFolder,"textures", j + ".png"),
                            os.path.join(self.tempfolder, "textures"),
                        )
                        temp += "newmtl " + j.split("/")[-1] + "\n"
                        temp += "Ka 1.000 1.000 1.000\n"
                        temp += "map_Kd " + os.path.join("textures", j.split("/")[-1] + ".png") + "\n"
                        temp += "\n"
        temp += "newmtl " + "missing" + "\n"
        temp += "Ka 1.000 1.000 1.000\n"
        temp += (
            "map_Kd " + os.path.join("textures", "Minecraft_missing_texture_block.svg.png") + "\n"
        )
        temp += "\n"
        shutil.copy(
            os.path.join(Path(__file__).parent.parent.parent, "resource","Minecraft_missing_texture_block.svg.png"),
            os.path.join(self.tempfolder, "textures", "Minecraft_missing_texture_block.svg.png"),
        )
        with open(os.path.join(self.tempfolder, self.name + ".mtl"), "w") as f:
            f.write(temp)

        oneblock = ""
        for blocks in self.tmpdata:
            for v in self.tmpdata[blocks]["v"]:
                if self.vtof.get((v[0], v[1], v[2])) == None:
                    oneblock += "v " + str(v[0]) + " " + str(v[1]) + " " + str(v[2]) + "\n"
                    self.vtof[(v[0], v[1], v[2])] = len(self.vtof) + 1
            for vt in self.tmpdata[blocks]["vt"]:
                for i in vt:
                    if self.vtovt.get((i[0], i[1])) == None:
                        self.vtovt[(i[0], i[1])] = len(self.vtovt) + 1
                        oneblock += "vt " + str(i[0]) + " " + str(i[1]) + "\n"

            ct1 = 0
            for f in self.tmpdata[blocks]["f"]:
                if "textures" in self.tmpdata[blocks]:
                    try:
                        oneblock += (
                            "usemtl " + self.tmpdata[blocks]["textures"][ct1].split("/")[-1] + "\n"
                        )
                    except:
                        oneblock += "usemtl " + "missing" + "\n"
                else:
                    oneblock += "usemtl " + "missing" + "\n"
                oneblock += "f "
                ct = 0
                for i in f:
                    oneblock += (
                        str(
                            self.vtof[
                                (
                                    self.tmpdata[blocks]["v"][i - 1][0],
                                    self.tmpdata[blocks]["v"][i - 1][1],
                                    self.tmpdata[blocks]["v"][i - 1][2],
                                )
                            ]
                        )
                        + "/"
                        + str(
                            self.vtovt[
                                (
                                    self.tmpdata[blocks]["vt"][ct1][ct][0],
                                    self.tmpdata[blocks]["vt"][ct1][ct][1],
                                )
                            ]
                        )
                        + " "
                    )
                    ct += 1
                oneblock += "\n"
                ct1 += 1
            oneblock += "#DEBUG \n"
        self.output += "mtllib " + self.name + ".mtl" + "\n"
        self.output += oneblock
        self.objfile.write(self.output)


if __name__ == "__main__":
    with open("./test.json", "r", encoding="utf8") as f:
        data = json.load(f)
    size = (
        int(data["Metadata"]["EnclosingSize"]["x"]),
        int(data["Metadata"]["EnclosingSize"]["y"]),
        int(data["Metadata"]["EnclosingSize"]["z"]),
    )
    data = data["Regions"]["Unnamed"]["decode_BlockStates"]
    Objhandel("test", data, size)

